//
// Uniport
//
// Treiber fuer universellen Portzugriff
//
// written 1998 by Guido Wischrop
//

#include "ntddk.h"              //Standard-Headerfile des DDK
#include "uniport.h"            //Headerfile mit Konstantendefinitionen


// Macro fuer die Debug-Ausgabe - Ausgabe erfolgt nur im "Checked-Build"
#if DBG
#define DbgPrint(arg) DbgPrint arg
#else
#define DbgPrint(arg)
#endif


//
// UniPort_DispatchCreateClose
//
// Die Funktion wird vom I/O Manager bei CreateFile('\\\\.\\UniPort',...)
// und CloseHandle aufgerufen.
// Da der Treiber keine Ressourcen belegt, wird nur der Status auf
// STATUS_SUCCESS gesetzt und die Anfrage mit IoCompleteRequest
// beendet.
// 

NTSTATUS UniPort_DispatchCreateClose(IN PDEVICE_OBJECT devObj,IN PIRP Irp)

{
    DbgPrint(("UniPort: Open / Close\n")); 
    Irp->IoStatus.Information = 0;
    Irp->IoStatus.Status = STATUS_SUCCESS;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    return STATUS_SUCCESS;
}


// UniPort_DispatchDeviceControl
//
// Die Funktion wird vom I/O Manager bei DeviceIOControl Aufrufen
// aktiviert. ber Parameters.DeviceIoControl.IoControlCode in der
// aktuellen IRP-Stackposition wird der Funktionscode ermittelt.
// Abhaengig von der Laenge des Input- bzw. des Output-Puffers
// wird der Portzugriff auf 1 Byte (CHAR), 2 Bytes (SHORT) oder
// 4 Bytes (DWORD) durchgefuehrt.
//

NTSTATUS UniPort_DispatchDeviceControl(IN PDEVICE_OBJECT devObj,IN PIRP Irp)

{

struct mParamBlockStruct
     {
      ULONG PortAdr;            // Portadresse
      
      ULONG Size;               // bestimmt die Groesse des Zugriffs
                                // 1: Read_Port_UChar
                                // 2: Read_Port_UShort
                                // 4: Read_Port_ULong
                                
      ULONG Value;              // nur bei Write - enthaelt den Wert,
                                // der auf den Port geschrieben werden soll
     } *mParamBlock;

    UCHAR *mBuffer;
    ULONG mInputBufferLength, mOutputBufferLength;
    ULONG mIOCTLCode;
    ULONG mPortAdr, mSize, mValue;
    NTSTATUS mStatus;
    
    PIO_STACK_LOCATION IrpStack = IoGetCurrentIrpStackLocation(Irp);

//  I/O Control Code lesen    
    mIOCTLCode=IrpStack->Parameters.DeviceIoControl.IoControlCode;

//  Laenge der Ein-/Ausgabepuffer lesen
    mInputBufferLength=IrpStack->Parameters.DeviceIoControl.InputBufferLength;
    mOutputBufferLength=IrpStack->Parameters.DeviceIoControl.OutputBufferLength;

    DbgPrint(("UniPort: InputBufferLength = %x\n",mInputBufferLength));
    DbgPrint(("UniPort: OutputBufferLength = %x\n",mOutputBufferLength));
    DbgPrint(("UniPort: FunctionCode = %x\n",mIOCTLCode));
    
//  Parameterblock auslesen
    mParamBlock=Irp->AssociatedIrp.SystemBuffer;
    mPortAdr=mParamBlock->PortAdr;
    mSize=mParamBlock->Size;

    if (mIOCTLCode==IOCTL_UniPort_ReadPort)
      {
        if (mOutputBufferLength<mSize)
          {
           Irp->IoStatus.Information = 0;
           mStatus = STATUS_BUFFER_TOO_SMALL;
          } else
          {
           Irp->IoStatus.Information = mSize;
           mStatus = STATUS_SUCCESS;
           mBuffer=Irp->AssociatedIrp.SystemBuffer;  
           switch (mSize)
                 {
                   case  1: *mBuffer=READ_PORT_UCHAR(mPortAdr);
                            break;
                   case  2: *mBuffer=READ_PORT_USHORT(mPortAdr);
                            break;
                   case  4: *mBuffer=READ_PORT_ULONG(mPortAdr);
                            break;
                   default: Irp->IoStatus.Information = 0;
                            mStatus = STATUS_INVALID_PARAMETER;
                 }
          }       
      } else
    if (mIOCTLCode==IOCTL_UniPort_WritePort)
      {
       Irp->IoStatus.Information = 0;
       mStatus = STATUS_SUCCESS;
       mValue=mParamBlock->Value;
       switch (mSize)
              {
                case  1: WRITE_PORT_UCHAR(mPortAdr, mValue);
                         break;
                case  2: WRITE_PORT_USHORT(mPortAdr, mValue);
                         break;
                case  4: WRITE_PORT_ULONG(mPortAdr, mValue);
                         break;
                default: mStatus = STATUS_INVALID_PARAMETER;
              }
      } else
      {
       Irp->IoStatus.Information = 0;
       mStatus = STATUS_NOT_SUPPORTED;
      }
 Irp->IoStatus.Status = mStatus;
 IoCompleteRequest(Irp, IO_NO_INCREMENT);
 return mStatus;
}

//
// UniPort_Unload
//
// Diese Funktion ermoeglicht das dynamische Entalden des Treibers.  
// Der symbolische Link und das Deviceobjekt werden entfernt.
//


VOID UniPort_Unload( IN PDRIVER_OBJECT DriverObject)
{
     WCHAR LinkName[]=L"\\DosDevices\\UniPort";
     UNICODE_STRING uniLinkName;
     NTSTATUS mStatus;     

     DbgPrint(("UniPort: Eintritt in Unload Routine !\n"));

     RtlInitUnicodeString(&uniLinkName, LinkName);
 
     mStatus=IoDeleteSymbolicLink(&uniLinkName);
     if (!NT_SUCCESS(mStatus))
      {
        DbgPrint(("UniPort: Kann Link nicht entfernen !\n"));
      } else
      {
       DbgPrint(("UniPort: Link entfernt !\n"));
       IoDeleteDevice(DriverObject->DeviceObject);
      } 
     DbgPrint(("UniPort: Unload Ende !\n"));
}


//
// DriverEntry
//
// Wird vom I/O Manager beim Starten des Treibers aufgerufen.
// Hier wird das Deviceobjekt und der symbolische Link erzeugt.
// Ausserdem werden im Treiberobjekt die Einsprungadressen der
// verfuegbaren Funktionen eingetragen.
// 

NTSTATUS DriverEntry(IN PDRIVER_OBJECT DriverObject,
                     IN PUNICODE_STRING RegistryPath)

{
     PDEVICE_OBJECT DeviceObject;
     NTSTATUS mStatus;
     WCHAR Name[]=L"\\Device\\UniPort";
     WCHAR DOSName[]=L"\\DosDevices\\UniPort";
     UNICODE_STRING uniName, uniDOSName;


//   Eintragen der Einsprungadressen in das Treiberobjekt     
     DriverObject->MajorFunction[IRP_MJ_CREATE]=UniPort_DispatchCreateClose;
     DriverObject->MajorFunction[IRP_MJ_CLOSE]=UniPort_DispatchCreateClose;
     DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL]=UniPort_DispatchDeviceControl;
     DriverObject->DriverUnload=UniPort_Unload;

//   wir benoetigen Unicode Strings, daher Umwandlung ANSI - Unicode
     RtlInitUnicodeString(&uniName, Name);
     RtlInitUnicodeString(&uniDOSName, DOSName);
     
//   das Geraereobjekt wird erzeugt
     mStatus=IoCreateDevice(DriverObject, 0, &uniName, FILE_DEVICE_UNKNOWN,
                            0, TRUE, &DeviceObject);

     if (!NT_SUCCESS(mStatus))
      {
        DbgPrint(("UniPort: Kann DeviceObject nicht erstellen !\n"));
        return mStatus;
      } 

//   der Symbolische Link wird erzeugt
     mStatus=IoCreateSymbolicLink(&uniDOSName, &uniName);
     if (!NT_SUCCESS(mStatus))
      {
        DbgPrint(("UniPort: Kann Link nicht erstellen !\n"));
        return mStatus;
      } else
      {
        DbgPrint(("UniPort: Link: %s - DosLink: %s\n", Name, DOSName));
      } 

//  der Treiber arbeitet mit Buffered IO
    DeviceObject->Flags |= DO_BUFFERED_IO;
    
    DbgPrint(("UniPort: Driver Entry erfolgreich abgearbeitet !\n"));
    return STATUS_SUCCESS;
}     
